DEFAULT_ALLOWED_TAGS = [
    "p",
    "div",
    "span",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "tt",
    "pre",
    "em",
    "strong",
    "ul",
    "sup",
    "li",
    "dl",
    "dd",
    "dt",
    "code",
    "img",
    "a",
    "table",
    "tr",
    "th",
    "td",
    "tbody",
    "caption",
    "colgroup",
    "thead",
    "tfoot",
    "blockquote",
    "ol",
    "hr",
    "br",
    "del",
    "ins",
]

DEFAULT_ALLOWED_ATTRIBUTES = {
    "*": [
        "class",
        "style",
        "id",
    ],
    "a": [
        "href",
        "target",
        "rel",
        "title",
    ],
    "img": [
        "src",
        "alt",
        "title",
    ],
    "tr": [
        "rowspan",
        "colspan",
    ],
    "td": [
        "rowspan",
        "colspan",
        "align",
    ],
}

DEFAULT_ALLOWED_STYLES = [
    "color",
    "background-color",
    "font-family",
    "font-weight",
    "font-size",
    "width",
    "height",
    "text-align",
    "border",
    "border-top",
    "border-bottom",
    "border-left",
    "border-right",
    "padding",
    "padding-top",
    "padding-bottom",
    "padding-left",
    "padding-right",
    "margin",
    "margin-top",
    "margin-bottom",
    "margin-left",
    "margin-right",
]

DEFAULT_BLEACH_KWARGS = {
    "tags": DEFAULT_ALLOWED_TAGS,
    "attributes": DEFAULT_ALLOWED_ATTRIBUTES,
    "styles": DEFAULT_ALLOWED_STYLES,
}


SETTINGS_MODE_OVERRIDE = "override"



from django import forms
from django.utils.functional import cached_property
from wagtail.blocks import TextBlock

from .utils import render_markdown
from .widgets import MarkdownTextarea


class MarkdownBlock(TextBlock):
    @cached_property
    def field(self):
        field_kwargs = {"widget": MarkdownTextarea(attrs={"rows": self.rows})}
        field_kwargs.update(self.field_options)
        return forms.CharField(**field_kwargs)

    def render_basic(self, value, context=None):
        return render_markdown(value, context)



import contextlib
import xml.etree.ElementTree as etree

from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from wagtail.images import get_image_model


# TODO: Default spec and class should be configurable, because they're
# dependent on how the project is set up.  Hard-coding of 'left',
# 'right' and 'full-width' should be removed.


class Linker:
    def run(self, fname, options):
        opts = {
            "spec": "width-500",
            "classname": "left",
        }

        for opt in options:
            bits = opt.split("=", 1)
            opt = bits[0]
            value = ""

            if len(bits) > 1:
                value = bits[1]

            if opt == "left":
                opts["classname"] = "left"
            elif opt == "right":
                opts["classname"] = "right"
            elif opt == "full":
                opts["classname"] = "full-width"
            elif opt == "width":
                with contextlib.suppress(ValueError):
                    opts["spec"] = f"width-{int(value)}"

        try:
            image = get_image_model().objects.get(title=fname)
        except ObjectDoesNotExist:
            return f'[image "{fname}" not found]'
        except MultipleObjectsReturned:
            return f'[multiple images "{fname}" found]'

        image_url = image.file.url
        rendition = image.get_rendition(opts["spec"])

        a = etree.Element("a")
        a.set("data-toggle", "lightbox")
        a.set("data-type", "image")
        a.set("href", image_url)
        img = etree.SubElement(a, "img")
        img.set("src", rendition.url)
        img.set("class", opts["classname"])
        img.set("width", str(rendition.width))
        img.set("height", str(rendition.height))
        return a



from importlib import import_module

from markdown import Extension
from markdown.inlinepatterns import Pattern


LINKER_RE = r"<:([a-z]+:)?([^>|\n]+)((\|[^>|\n]+){0,})>"


class LinkerPattern(Pattern):
    def __init__(self, pattern, md, link_types):
        super().__init__(pattern, md)
        self.link_types = link_types

    def handleMatch(self, m):
        link_types = self.link_types
        opts = []
        if m.group(3) is not None and len(m.group(4)):
            opts = m.group(4).split("|")[1:]

        link_type = m.group(2)
        if link_type is None:
            link_type = "__default__"

        try:
            mod = import_module(link_types[link_type])
            c = mod.Linker()

            return c.run(m.group(3), opts)
        except KeyError:
            return f'[invalid linker type "{link_type.rstrip(":")}"]'


class LinkerExtension(Extension):
    def __init__(self, link_types, **kwargs):
        super().__init__(**kwargs)
        self.link_types = link_types

    def extendMarkdown(self, md):
        md.inlinePatterns.register(
            LinkerPattern(LINKER_RE, md, self.link_types), "linker", 9
        )


def makeExtension(**kwargs):
    """As per https://python-markdown.github.io/extensions/api/#dot_notation"""
    return LinkerExtension(**kwargs)  # pragma: no cover



from django import forms
from django.conf import settings
from wagtail.admin.staticfiles import versioned_static


class MarkdownTextarea(forms.Textarea):
    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)
        attrs["data-controller"] = "easymde"

        if "autodownload_fontawesome" in getattr(settings, "WAGTAILMARKDOWN", {}):
            autodownload = (
                "true"
                if settings.WAGTAILMARKDOWN["autodownload_fontawesome"]
                else "false"
            )
            attrs["data-easymde-autodownload-value"] = autodownload

        return attrs

    @property
    def media(self):
        return forms.Media(
            css={
                "all": (
                    versioned_static("wagtailmarkdown/css/easymde.min.css"),
                    versioned_static("wagtailmarkdown/css/easymde.tweaks.css"),
                )
            },
            js=(
                versioned_static("wagtailmarkdown/js/easymde.min.js"),
                versioned_static("wagtailmarkdown/js/easymde.attach.js"),
                versioned_static("wagtailmarkdown/js/easymde-controller.js"),
            ),
        )



from collections import defaultdict

import bleach
import markdown

from django.conf import settings
from django.utils.encoding import smart_str
from django.utils.safestring import mark_safe

from wagtailmarkdown.constants import DEFAULT_BLEACH_KWARGS, SETTINGS_MODE_OVERRIDE
from wagtailmarkdown.mdx.inlinepatterns import (
    DelExtension,
    ImageExtension,
    LinkExtension,
)
from wagtailmarkdown.mdx.linker import LinkerExtension


def render_markdown(text, context=None):
    """
    Turn markdown into HTML.
    """
    markdown_html = _transform_markdown_into_html(text)
    sanitised_markdown_html = _sanitise_markdown_html(markdown_html)
    # note: we use mark_safe here because bleach is already sanitising the HTML
    return mark_safe(sanitised_markdown_html)  # noqa: S308


def _transform_markdown_into_html(text):
    return markdown.markdown(smart_str(text), **_get_markdown_kwargs())


def _sanitise_markdown_html(markdown_html):
    return bleach.clean(markdown_html, **_get_bleach_kwargs())


def _get_bleach_kwargs():
    bleach_kwargs = DEFAULT_BLEACH_KWARGS.copy()

    if not hasattr(settings, "WAGTAILMARKDOWN"):
        return bleach_kwargs

    override = (
        settings.WAGTAILMARKDOWN.get("allowed_settings_mode", "extend").lower()
        == SETTINGS_MODE_OVERRIDE
    )
    if "allowed_styles" in settings.WAGTAILMARKDOWN:
        if override:
            bleach_kwargs["styles"] = settings.WAGTAILMARKDOWN["allowed_styles"]
        else:
            bleach_kwargs["styles"] = list(
                set(
                    settings.WAGTAILMARKDOWN["allowed_styles"] + bleach_kwargs["styles"]
                )
            )
    if "allowed_tags" in settings.WAGTAILMARKDOWN:
        if override:
            bleach_kwargs["tags"] = settings.WAGTAILMARKDOWN["allowed_tags"]
        else:
            bleach_kwargs["tags"] = list(
                set(settings.WAGTAILMARKDOWN["allowed_tags"] + bleach_kwargs["tags"])
            )

    if "allowed_attributes" in settings.WAGTAILMARKDOWN:
        if override:
            bleach_kwargs["attributes"] = settings.WAGTAILMARKDOWN["allowed_attributes"]
        else:
            merged = defaultdict(set)
            for _dict in [
                bleach_kwargs["attributes"],
                settings.WAGTAILMARKDOWN["allowed_attributes"],
            ]:
                for key, value in _dict.items():
                    merged[key].update(value)
            bleach_kwargs["attributes"] = {
                key: list(value) for key, value in merged.items()
            }

    return bleach_kwargs


def _get_default_markdown_kwargs():
    kwargs = {
        "extensions": [
            "extra",
            "codehilite",
            "tables",
            LinkerExtension(
                {
                    "__default__": "wagtailmarkdown.mdx.linkers.page",
                    "page:": "wagtailmarkdown.mdx.linkers.page",
                    "image:": "wagtailmarkdown.mdx.linkers.image",
                    "doc:": "wagtailmarkdown.mdx.linkers.document",
                }
            ),
            LinkExtension(),
            ImageExtension(),
            DelExtension(),
        ],
        "extension_configs": {
            "codehilite": [
                ("guess_lang", False),
            ]
        },
        "output_format": "html5",
    }

    return kwargs


def _get_markdown_kwargs():
    kwargs = _get_default_markdown_kwargs()

    if not hasattr(settings, "WAGTAILMARKDOWN"):
        return kwargs

    markdown_settings = settings.WAGTAILMARKDOWN
    override = (
        markdown_settings.get("extensions_settings_mode", "extend").lower()
        == SETTINGS_MODE_OVERRIDE
    )
    if "extensions" in markdown_settings:
        if override:
            kwargs["extensions"] = markdown_settings["extensions"]
        else:
            kwargs["extensions"] = list(
                set(kwargs["extensions"] + markdown_settings["extensions"])
            )

    if "extension_configs" in markdown_settings:
        if override:
            kwargs["extension_configs"] = markdown_settings["extension_configs"]
        else:
            kwargs["extension_configs"].update(markdown_settings["extension_configs"])

    if "tab_length" in markdown_settings:
        kwargs["tab_length"] = markdown_settings["tab_length"]

    return kwargs
